/**
//  \file 3D03-GetFramePeaknStatus.cpp
//	
//  \brief
//  This example is an application to get familar with the 3D camera MV1-D2048X1088-3D03-760-G2
//  by getting 3D data and the useful status information (image counter, real time counter, 
//  encoder position, missed trigger counter, etc.) contained in the raw frame. 
//  
//  Description: The main functionality of the example is to configure the camera with 
//  a set of properties and grab some frames. 
//  The example shows how to capture the raw data coming from the camera (2D and 3D) and 
//  how to store them. This data contains not only the laser peak coordinates, the laser  
//  line width and height but also some status information specific for each frame: 
//  image counter, real time counter in microsec (time stamp), encoder position,
//  missed trigger counter, etc. The user may use this raw data for his own purposes,
//  for instance, in order to identify frames, to count the lost frames or to sinchronize
//  two cameras.
//
//  The properties setup is implemented in the "Configure" function.
//  The property list is available in the documentation of the PFRemote.
//
//  The main purpose of the "AcquireImages" function is to grab some raw frames from
//  the framegrabber and store the information. The number of frames to grab is defined
//  by MAXPROFILES.
//  Depending on the PeakDetector_Mode property, the information contained in
//  the frames is different:
//  - In "Mode_2Donly" laser detection is turned off and camera behaves as a 
//  normal area scan camera.
//
//  - In "Mode2Dand3D" laser line detection is turned on. The sensor image (2D image) 
//  is transmitted together with the 3D data. For the laser line detection there are 
//  4 additional lines that contain the 3D data, including the laser peak coordinate,
//  the laser line width and height (highest grey value inside the laser line). 
//  Find more information in the section "3D data format" of the 3D02 camera manual.
//
//  - In "Mode_3Donly" laser line detection is turned on and only 3D data plus an 
//  additional image row is transmitted. The scan rate of this mode is considerably 
//  faster than the 2D&3D mode.
//
//  In the case of Mode_2Donly or Mode2Dand3D, this example copies all the pixels of
//  a raw frame to a memory buffer and then stores this buffer into a raw image file.
//  This kind of raw files have no format but you can visualize them with a viewer
//  like "ImageJ".
//
//  For the case of Mode2Dand3D or Mode_3Donly, this example shows how the 3D data 
//  is transmitted from the camera and how to capture the laser peak coordinate, 
//  the laser line width and height and store them into a txt file. 
// 
*/

#ifdef WIN32
#include <windows.h>
#include <conio.h>
#include <process.h>
#include <direct.h>
#endif // WIN32

#include <string.h>
#include <iostream>
#include <fstream>
#include <stdio.h>
#include <stdlib.h>

using namespace std;

#ifdef _UNIX_

#include <termios.h>
#include <unistd.h>
#include <fcntl.h>

int _kbhit(void)
{
  struct termios oldt, newt;
  int ch;
  int oldf;

  tcgetattr(STDIN_FILENO, &oldt);
  newt = oldt;
  newt.c_lflag &= ~(ICANON | ECHO);
  tcsetattr(STDIN_FILENO, TCSANOW, &newt);
  oldf = fcntl(STDIN_FILENO, F_GETFL, 0);
  fcntl(STDIN_FILENO, F_SETFL, oldf | O_NONBLOCK);

  ch = getchar();

  tcsetattr(STDIN_FILENO, TCSANOW, &oldt);
  fcntl(STDIN_FILENO, F_SETFL, oldf);

  if(ch != EOF)
  {
    ungetc(ch, stdin);
    return 1;
  }

  return 0;
}

#define _getch getchar

#endif // _UNIX_

#include <PvDeviceFinderWnd.h>
#include <PvDevice.h>
#include <PvPipeline.h>
#include <PvBuffer.h>
#include <PvStream.h>
#include <PvStreamRaw.h>

#define MAXPROFILES 200
#define MAX_PATH 256


unsigned char *UserBuffer;
char gsFullAppPath[MAX_PATH];


void Configure(PvGenParameterArray *lDeviceParams)
{
    PvResult lResult;

    // Setup the camera configuration

    // 1- Set the Width and OffsetX properties
    // Property: Width - equivalent to Window.W
    // Parameter: Width of the Window 
    // (Min:32 Max:2048, in increments of 8 pixels)
    lResult = lDeviceParams->SetIntegerValue( "Width", 1600 );
    if(!lResult.IsOK())	printf("Set camera property failed: Width");
     // Property: OffsetX - equivalent to Window.X
    // (Min:0 Max:448, in increments of 8 pixels)
    lResult = lDeviceParams->SetIntegerValue( "OffsetX", 200 );
    if(!lResult.IsOK())	printf("Set camera property failed: OffsetX");

    // 2- Set the Peak0 ROI
    // Property: Peak0_3DY- equivalent to Window.Y
    // (Min:0 Max:1072, in increments of 1 pixel)
    lResult = lDeviceParams->SetIntegerValue( "Peak0_3DY", 100 );
    if(!lResult.IsOK())	printf("Set camera property failed: Peak0_3DY");
     // Property: Peak0_3DH- equivalent to Window.H
    // (Min:16 Max:1009, in increments of 1 pixel)
    lResult = lDeviceParams->SetIntegerValue( "Peak0_3DH", 1000 );
    if(!lResult.IsOK())	printf("Set camera property failed: Peak0_3DH");

    // 3- Set the number of peaks. In this case only 1
    // Property: PeakDetector_NrOfPeaks (1 or 2)
    lResult = lDeviceParams->SetIntegerValue("PeakDetector_NrOfPeaks", 1);
    if(!lResult.IsOK())	printf("Set camera property failed: PeakDetector_NrOfPeaks");

    // 4- Set the 3D mode
    // Property: PeakDetector_Mode	
    // Parameter: 0: "Img2D" (2D only)
    //            1: "Img2D3D" (2D+3D)
    //            2: "Img3D" (3D only)
    lResult = lDeviceParams->SetEnumValue("PeakDetector_Mode", 1);
    if(!lResult.IsOK())	printf("Set camera property failed: PeakDetector_Mode");

    // 5- Read the HeightInterface property and set the Height property according to it
    PvInt64 iProperty = 0;
    lResult = lDeviceParams->GetIntegerValue("HeightInterface",iProperty);
    lResult = lDeviceParams->SetIntegerValue("Height", iProperty);
    if(!lResult.IsOK())	printf("Set camera property failed: Height");

    // 6- Set the Trigger properties

    // Property: TriggerMode
    // External trigger mode. TriggerMode=Off means Free-running trigger.
    // Parameter: 0 : Off
    //            1 : On
    lResult = lDeviceParams->SetEnumValue("TriggerMode", 0);
    if(!lResult.IsOK())	printf("Set camera property failed: TriggerMode");

    // Property: TriggerSource
    // Parameter: 0  : Software: TriggerSoftware as trigger source.
    //            2  : Line1: ISO_IN1 as trigger source
    //            25 : PLC_Q4: PLC_Q4 output as trigger source.
    //            32 : ABTrigger: Differential input trigger.   
    lResult = lDeviceParams->SetEnumValue("TriggerSource", 0);
    if(!lResult.IsOK())	printf("Set camera property failed: TriggerSource");

    // Property: TriggerDivider
    // Parameter: TriggerDivider value - only take every n-th trigger (n=TriggerDivider)
    // TriggerDivider=1 means: take every trigger (default behaviour)
    lResult = lDeviceParams->SetIntegerValue("TriggerDivider", 1);
    if(!lResult.IsOK())	printf("Set camera property failed: TriggerDivider");

    // 7- Set the exposure time

    // Property: ExposureTime
    // Parameter: Exposure time of the camera in millisecond. Maximum value is 419ms.
    double f=(double)100.0;
    lResult = lDeviceParams->SetFloatValue("ExposureTime", f);
    if(!lResult.IsOK())	printf("Set camera property failed: ExposureTime");

    // 8- Set the other properties

    int iThreshold=210;
    // Property:Peak0_Threshold
    // Peak detector gray-level threshold (pre-filter threshold)
    // (Min:0 Max:2047)
    lResult = lDeviceParams->SetIntegerValue("Peak0_Threshold", iThreshold);
    if(!lResult.IsOK())	printf("Set camera property failed: Peak0_Threshold");
    // Property:Peak1_Threshold
    // Peak detector gray-level threshold (pre-filter threshold)
    // (Min:0 Max:2047)
    lResult = lDeviceParams->SetIntegerValue("Peak1_Threshold", iThreshold);
    if(!lResult.IsOK())	printf("Set camera property failed: Peak1_Threshold");


    // Feel free to add some more properties...
    
    // Read the properties to check the desired value
    lDeviceParams->GetEnumValue("PeakDetector_Mode",iProperty);
    lDeviceParams->GetIntegerValue("Width",iProperty);
    lDeviceParams->GetIntegerValue("Height",iProperty);
    lDeviceParams->GetEnumValue("TriggerSource",iProperty);
    lDeviceParams->GetFloatValue("ExposureTime",f);
}

void ProcessFrame(int *iProfilecount, int PeakDetectorMode, int Width, int Height, FILE * pPeakFile, FILE * pStatusFile)
{
    char text[90];
    FILE * pFile;
    unsigned char *pixel;
    int numcar;

    // Status information variables 
    unsigned int ImageCounter = 0;       // 3 bytes
    unsigned int RealTimeCounter = 0;    // 4 bytes
    unsigned int EncoderPosition = 0;    // 3 bytes
    unsigned int MissedTriggerCounter = 0;   // 1 byte
    unsigned int MissedBurstTriggerCounter = 0;   // 1 byte
    unsigned int MissedFCTriggerCounter = 0;   // 1 byte
    unsigned char halfbyteorder = 0;
    unsigned int stat = 0;

    if(PeakDetectorMode==0) //2Donly mode
    {
        sprintf(text,"./Frames/image%04d.raw", (*iProfilecount)+1);
        pFile = fopen (text, "w");
        if (pFile == NULL) perror ("Error opening file");
        else 
        {
            fwrite(UserBuffer, 1, Width*Height, pFile);
            fclose (pFile);
        }
    }
    if(PeakDetectorMode==1) //2Dand3D mode
    {
        sprintf(text, "./Frames/image%04d.raw", (*iProfilecount)+1);
        pFile = fopen (text, "w");
        if (pFile == NULL) perror ("Error opening file");
        else 
        {
            fwrite(UserBuffer, 1, Width*(Height-5), pFile);
            fclose (pFile);
        }
    }
    if((PeakDetectorMode==1) || (PeakDetectorMode==2)) //2Dand3D mode or 3Donly mode
    {
        // The 3D data is located at the end of each frame. In the 3D camera MV1-D2048X1088-3D03-760-G2
        // the 3D Data Block consists of 5 lines. As the Row 0 of the 3D Data Block contains 2D image 
        // information (see below), the begining of the profile will be at ((Height-4) * Width)
        unsigned char *startProfilePixel = UserBuffer + ((Height-4) * Width);
        for(int x=0; x < Width; x++)
        {
            // 3D Data Block: For each column there are 5 lines that contain the 3D data.
            // More information in the section "3D data format" of the 3D03 camera manual
            // Row 0: 2D Image information
            // Row 1: Peak[15:8]
            // Row 2: Peak[7:0] -> Note: Peak[15:6] Integer part: Peak[5:0] Fractional Part
            // Row 3: LL_WIDTH [5:0] Laser Line Width
            // Row 4: LL_HEIGHT [8:4] Laser Line Height; STAT[3:0] Status information
            pixel = startProfilePixel + x;
            // Get the peak value in float format
            //float fPeak = ((p.pixel(x,y)<<8) + p.pixel(x,y+1))/64;
            float fpeak = ((*pixel<<8) + *(pixel+Width))/64;
            // Get the peak value in 8bit format
            unsigned char cpeak= *pixel;
            unsigned char laserw = *(pixel+(2*Width));
            unsigned char laserq = *(pixel+(3*Width)) >> 4;
            // Save the 3D data into a txt file
            numcar = sprintf(text,"Profile:%d Col:%d Peak(float):%f Peak(8bits):%d Width:%d Quality:%d\n",
                *iProfilecount, x, fpeak, cpeak, laserw, laserq);
            fwrite(text, 1, numcar, pPeakFile);

            pixel = pixel + (3*Width);
            // Get the corresponding part of the status information line
            // More information in the section "Status information" of the 3D03 camera manual
            if(x<6) //ImageCounter: 3 bytes (6 = 3 * 2 halfbytes)
            {
                stat = (unsigned int)(*pixel & 0x0f)<<(4*halfbyteorder);
                ImageCounter = ImageCounter + stat;
                halfbyteorder++;
                if(halfbyteorder==6) halfbyteorder = 0;
            }
            if(x>=8 && x<16) //RealTimeCounter: 4 bytes
            {
                stat = (unsigned int)(*pixel & 0x0f)<<(4*halfbyteorder);
                RealTimeCounter = RealTimeCounter + stat;
                halfbyteorder++;
                if(halfbyteorder==8) halfbyteorder = 0;
            }
            if(x>=16 && x<22) //EncoderPosition: 3 bytes
            {
                stat = (unsigned int)(*pixel & 0x0f)<<(4*halfbyteorder);
                EncoderPosition = EncoderPosition + stat;
                halfbyteorder++;
                if(halfbyteorder==6) halfbyteorder = 0;
            }
            if(x>=24 && x<26) //MissedTriggerCounter: 1 byte
            {
                stat = (unsigned int)(*pixel & 0x0f)<<(4*halfbyteorder);
                MissedTriggerCounter = MissedTriggerCounter + stat;
                halfbyteorder++;
                if(halfbyteorder==2) halfbyteorder = 0;
            }
            if(x>=32 && x<34) //MissedBurstTriggerCounter: 1 byte
            {
                stat = (unsigned int)(*pixel & 0x0f)<<(4*halfbyteorder);
                MissedBurstTriggerCounter = MissedBurstTriggerCounter + stat;
                halfbyteorder++;
                if(halfbyteorder==2) halfbyteorder = 0;
            }
            if(x>=40 && x<42) //MissedFrameCombineTriggerCounter: 1 byte
            {
                stat = (unsigned int)(*pixel & 0x0f)<<(4*halfbyteorder);
                MissedFCTriggerCounter = MissedFCTriggerCounter + stat;
                halfbyteorder++;
                if(halfbyteorder==2) halfbyteorder = 0;
            }
        }
        // Save the status info into a file
        numcar = sprintf(text,"Profile:%d ImageCounter:%u RealTimeCounter:%u EncoderPosition:%u",
                *iProfilecount, ImageCounter, RealTimeCounter, EncoderPosition);
        fwrite(text, 1, numcar, pStatusFile);
        numcar = sprintf(text,"  MissedTriggerCounter:%u MissedBurstTriggerCounter:%u MissedFCTriggerCounter:%u\n",
                MissedTriggerCounter, MissedBurstTriggerCounter, MissedFCTriggerCounter);
        fwrite(text, 1, numcar, pStatusFile);

        *iProfilecount = *iProfilecount + 1;
    }
}

bool AcquireImages()
{
    // Create a GEV Device finder dialog
    PvDeviceFinderWnd lDeviceFinderWnd;

    // Prompt the user to select a GEV Device
    lDeviceFinderWnd.ShowModal();

    // Get the connectivity information for the selected GEV Device
    PvDeviceInfo* lDeviceInfo = lDeviceFinderWnd.GetSelected();

    // If no device is selected, abort
    if( lDeviceInfo == NULL )
    {
        printf( "No device selected.\n" );
        return false;
    }

    // Connect to the GEV Device
    PvDevice lDevice;
    printf( "Connecting to %s\n", lDeviceInfo->GetMACAddress().GetAscii() );
    if ( !lDevice.Connect( lDeviceInfo ).IsOK() )
    {
        printf( "Unable to connect to %s\n", lDeviceInfo->GetMACAddress().GetAscii() );
        return false;
    }
    printf( "Successfully connected to %s\n", lDeviceInfo->GetMACAddress().GetAscii() );
    printf( "\n" );

    // Get device parameters need to control streaming
    PvGenParameterArray *lDeviceParams = lDevice.GetGenParameters();

    // Camera configuration
    Configure(lDeviceParams);  

    // Read PeakDetectorMode
    PvInt64 PeakDetectorMode;
    lDeviceParams->GetEnumValue("PeakDetector_Mode",PeakDetectorMode);

    // Negotiate streaming packet size
    lDevice.NegotiatePacketSize();

    // Create the PvStream object
    PvStream lStream;

    // Open stream - have the PvDevice do it for us
    printf( "\r\nOpening stream to device\n" );
    lStream.Open( lDeviceInfo->GetIPAddress() );

    // Create the PvPipeline object
    PvPipeline lPipeline( &lStream );
    
    // Reading payload size from device
    PvInt64 lSize = 0;
	lDeviceParams->GetIntegerValue( "PayloadSize", lSize );

    // Set the Buffer size and the Buffer count
    lPipeline.SetBufferSize( static_cast<PvUInt32>( lSize ) );
    lPipeline.SetBufferCount( 16 ); // Increase for high frame rate without missing block IDs
    // Set the UserBuffer size
    UserBuffer = (unsigned char*)malloc(lSize);

    // Have to set the Device IP destination to the Stream
    lDevice.SetStreamDestination( lStream.GetLocalIPAddress(), lStream.GetLocalPort() );

    // IMPORTANT: the pipeline needs to be "armed", or started before 
    // we instruct the device to send us images
    printf( "Starting pipeline\n" );
    lPipeline.Start();

    // Get stream parameters/stats
    PvGenParameterArray *lStreamParams = lStream.GetParameters();

    // TLParamsLocked is optional but when present, it MUST be set to 1
    // before sending the AcquisitionStart command
	lDeviceParams->SetIntegerValue( "TLParamsLocked", 1 );

    printf( "Resetting timestamp counter...\n" );
	lDeviceParams->ExecuteCommand( "GevTimestampControlReset" );

    // The pipeline is already "armed", we just have to tell the device
    // to start sending us images
    printf( "Sending StartAcquisition command to device\n" );
	lDeviceParams->ExecuteCommand( "AcquisitionStart" );

    char lDoodle[] = "|\\-|-/";
    int lDoodleIndex = 0;
    PvInt64 lImageCountVal = 0;
    double lFrameRateVal = 0.0;
    double lBandwidthVal = 0.0;

    // Prepare the file streams to store the 3D data in a txt file
    getcwd(gsFullAppPath, MAX_PATH);
    string strPathPeakInfo(gsFullAppPath);
    strPathPeakInfo.append("/PeakInfo.txt");
    string strPathStatusInfo(gsFullAppPath);
    strPathStatusInfo.append("/StatusInfo.txt");
    //ofstream info_file(strPathInfo.c_str());
    FILE * pPeakFile;
    pPeakFile = fopen (strPathPeakInfo.c_str(), "w");
    if (pPeakFile == NULL) perror ("Error opening file");
    FILE * pStatusFile;
    pStatusFile = fopen (strPathStatusInfo.c_str(), "w");
    if (pStatusFile == NULL) perror ("Error opening file");

    int iProfilecount = 0;

    // Grab profiles until MAXPROFILES reached or the user instructs us to stop
    printf( "\n<press the enter key to stop streaming>\n" );
    while((iProfilecount < MAXPROFILES) && ( !_kbhit() ))
    {
        // Retrieve next buffer		
        PvBuffer *lBuffer = NULL;
        PvResult  lOperationResult;
        PvResult lResult = lPipeline.RetrieveNextBuffer( &lBuffer, 1000, &lOperationResult );
		PvUInt32 lWidth = 0, lHeight = 0;
        
        if ( lResult.IsOK() )
        {
            if ( lOperationResult.IsOK() )
            {
				lStreamParams->GetIntegerValue( "ImagesCount", lImageCountVal );
				lStreamParams->GetFloatValue( "AcquisitionRateAverage", lFrameRateVal );
				lStreamParams->GetFloatValue( "BandwidthAverage", lBandwidthVal );
            
				// If the buffer contains an image, display width and height
				
				if ( lBuffer->GetPayloadType() == PvPayloadTypeImage )
				{
					// Get image specific buffer interface
					PvImage *lImage = lBuffer->GetImage();
					// Read width, height
					lWidth = lBuffer->GetImage()->GetWidth();
					lHeight = lBuffer->GetImage()->GetHeight();
				}

				printf( "%c Timestamp: %012llX BlockID: %06llX W: %i H: %i %.01f FPS %.01f Mb/s\r", 
                    lDoodle[ lDoodleIndex ],
                    lBuffer->GetTimestamp(),
                    lBuffer->GetBlockID(),
					lWidth,
					lHeight,
                    lFrameRateVal,
                    lBandwidthVal / 1000000.0 ); 
            }
			// We have an image - do some processing (...) 
			PvUInt8 *img;
			img = NULL;
			img = lBuffer->GetDataPointer();
			memcpy(UserBuffer, img, lWidth*lHeight); // lWidth*lHeight is equivalent to lSize

            ProcessFrame(&iProfilecount, PeakDetectorMode, lWidth, lHeight, pPeakFile, pStatusFile);

			//and VERY IMPORTANT, release the buffer back to the pipeline
            lPipeline.ReleaseBuffer( lBuffer );
        }
        else
        {
            // Timeout
            printf( "%c Timeout\r", lDoodle[ lDoodleIndex ] );
        }

        ++lDoodleIndex %= 6;
    }

    if (_kbhit()) _getch(); // Flush key buffer for next stop
    printf( "\n\n" );

    fclose (pPeakFile);
    fclose (pStatusFile);
    
    // Tell the device to stop sending images
    printf( "Sending AcquisitionStop command to the device\n" );
	lDeviceParams->ExecuteCommand( "AcquisitionStop" );

    // If present reset TLParamsLocked to 0. Must be done AFTER the 
    // streaming has been stopped
	lDeviceParams->SetIntegerValue( "TLParamsLocked", 0 );

    // We stop the pipeline - letting the object lapse out of 
    // scope would have had the destructor do the same, but we do it anyway
    printf( "Stop pipeline\n" );
    lPipeline.Stop();

    // Now close the stream. Also optionnal but nice to have
    printf( "Closing stream\n" );
    lStream.Close();

    // Finally disconnect the device. Optional, still nice to have
    printf( "Disconnecting device\n" );
    lDevice.Disconnect();    

    return true;
}

//
// Main function
//

int main()
{
    // PvPipeline used to acquire images from a device
    printf( "Config and Grab sample - image acquisition from a device\n\n" );
    AcquireImages();

    printf( "\n<press the enter key to exit>\n" );
    while ( !_kbhit() );

    return 0;
}
