/**
//  \file 3D03-GetPeaknStatus.cpp
//	
//  \brief
//  This example is an application to get 3D data and the useful status information 
//  (image counter, real time counter, encoder position, missed trigger counter, etc.) 
//  with the 3D camera MV1-D2048X1088-3D03-760-G2 by using the information contained 
//  in the raw frame. 
//  
//  In this example no range map is created. Instead of that, the user can treat the
//  peak data by himself. 
//  See the 3D03-GetRangeMap example to create a range map from the grabbed profiles.
//
//  Description: The main functionality of the example is to configure the camera with 
//  a set of properties and grab some frames. 
//  The example shows how to capture the raw data coming from the camera and how to
//  store them. This data contains not only the laser peak coordinates, the laser line 
//  width and height but also some status information specific for each frame: 
//  image counter, real time counter in microsec (time stamp), encoder position,
//  missed trigger counter, etc. The user may use this raw data for his own purposes,
//  for instance, in order to identify frames, to count the lost frames or to sinchronize
//  two cameras.
//
//  The properties setup is implemented in the "config" function.
//  The property list is available in the documentation of the PFRemote.
//
//  The main purpose of the "capture" function is to grab some raw frames from
//  the framegrabber and store the information.
//  Depending on the PeakDetector_Mode property, the information contained in
//  the frames is different:
//  - In "Mode_2Donly" laser detection is turned off and camera behaves as a 
//  normal area scan camera.
//
//  - In "Mode2Dand3D" laser line detection is turned on. The sensor image (2D image) 
//  is transmitted together with the 3D data. For the laser line detection there are 
//  4 additional lines that contain the 3D data, including the laser peak coordinate,
//  the laser line width and height (highest grey value inside the laser line). 
//  Find more information in the section "3D data format" of the 3D02 camera manual.
//
//  - In "Mode_3Donly" laser line detection is turned on and only 3D data plus an 
//  additional image row is transmitted. The scan rate of this mode is considerably 
//  faster than the 2D&3D mode.
//
//  For this example the recommended mode is the Mode_3Donly, being the Mode2Dand3D 
//  also a valid but slower option. The Mode_2Donly is not suitable for this example.
// 
*/

#include <iostream>
#include <fstream>

#ifdef WIN32
#include <direct.h>
#include <conio.h>
#endif // WIN32

#ifdef _UNIX_

#include <termios.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <dlfcn.h>
#include <string.h>

int _kbhit(void)
{
  struct termios oldt, newt;
  int ch;
  int oldf;

  tcgetattr(STDIN_FILENO, &oldt);
  newt = oldt;
  newt.c_lflag &= ~(ICANON | ECHO);
  tcsetattr(STDIN_FILENO, TCSANOW, &newt);
  oldf = fcntl(STDIN_FILENO, F_GETFL, 0);
  fcntl(STDIN_FILENO, F_SETFL, oldf | O_NONBLOCK);

  ch = getchar();

  tcsetattr(STDIN_FILENO, TCSANOW, &oldt);
  fcntl(STDIN_FILENO, F_SETFL, oldf);

  if(ch != EOF)
  {
    ungetc(ch, stdin);
    return 1;
  }

  return 0;
}

#define _getch getchar
#define _getcwd getcwd
#define MAX_PATH 256

#endif // _UNIX_

#include <pf3dcam/Camera.hpp>

/// \brief
/// Maximum number of frames (profiles) to be captured
///
#define MAXPROFILES 200

/// \brief
/// Timeout (in milliseconds) to receive a valid frame from the camera
///
#define TIMEOUT 1000

#define CURRENTFRAME true
#define GETNEWFRAME false

using namespace std;
using namespace sal3d;
using namespace pf3dcam;

namespace
{
    char gsFullAppPath[MAX_PATH];
}

///
/// \brief To configure some properties of the camera
///
/// \return Error
///
int config(pf3dcam::Camera *_camera)
{
    //Inicialization for 3D03 camera only

    // Check the camera type
    CAM_Type _camType=CAM_UNKNOWN;
    _camType = _camera->type();
    if (_camType!=CAM_3D03)
        throw std::runtime_error ("This is not a 3D03 camera. This example is only valid for 3D03 cameras");

    // Setup the camera configuration

    // 1- Set the Width and OffsetX properties
    // Property: Width - equivalent to Window.W
    // Parameter: Width of the Window 
    // (Min:32 Max:2048, in increments of 8 pixels)
    if(_camera->SetCameraPropertyI("Width", 1600) < 0)
        throw std::runtime_error ("Set camera property failed: Width");
    // Property: OffsetX - equivalent to Window.X
    // (Min:0 Max:448, in increments of 8 pixels)
    if(_camera->SetCameraPropertyI("OffsetX", 200) < 0)
        throw std::runtime_error ("Set camera property failed: OffsetX");

    // 2- Set the Peak0 ROI
    // Property: Peak0_3DY- equivalent to Window.Y
    // (Min:0 Max:1072, in increments of 1 pixel)
    if(_camera->SetCameraPropertyI("Peak0_3DY", 100) < 0)
        throw std::runtime_error ("Set camera property failed: Peak0_3DY");
    // Property: Peak0_3DH- equivalent to Window.H
    // (Min:16 Max:1009, in increments of 1 pixel)
    if(_camera->SetCameraPropertyI("Peak0_3DH", 1000) < 0)
        throw std::runtime_error ("Set camera property failed: Peak0_3DH");

    // 3- Set the number of peaks. In this case only 1
    // Property: PeakDetector_NrOfPeaks (1 or 2)
    if(_camera->SetCameraPropertyI("PeakDetector_NrOfPeaks", 1) < 0)
        throw std::runtime_error ("Set camera property failed: PeakDetector_NrOfPeaks");

    // 4- Set the 3D mode
    // Property: PeakDetector.Mode	
    // Parameter: 0: "Img2D" (2D only)
    //            1: "Img2D3D" (2D+3D)
    //            2: "Img3D" (3D only)
    if(_camera->SetCameraPropertyI("PeakDetector.Mode", 2) < 0)
        throw std::runtime_error ("Set camera property failed: PeakDetector.Mode");

    // 5- Read the HeightInterface property and set the Height property according to it
    int iProperty;
    _camera->GetCameraPropertyI("HeightInterface",&iProperty);
    if(_camera->SetCameraPropertyI("Height", iProperty) < 0)
        throw std::runtime_error ("Set camera property failed: Height");

   // 6- Set the Trigger properties

    // Property: Trigger.Source
    // Parameter: 0 : Free_Running (Trigger is on Free Running Mode)
    //            1 : Interface_Trigger (Interface Trigger is the trigger source)
    //            2 : IO_Trigger (I/O Trigger is the trigger source)
    if(_camera->SetCameraPropertyI("Trigger.Source", 0) < 0)
        throw std::runtime_error ("Set camera property failed: Trigger.Source");

    // Property: TriggerDivider
    // Parameter: TriggerDivider value - only take every n-th trigger (n=TriggerDivider)
    // TriggerDivider=1 means: take every trigger (default behaviour)
    if(_camera->SetCameraPropertyI("TriggerDivider", 1) < 0)
        throw std::runtime_error ("Set camera property failed: TriggerDivider");

    // 7- Set the exposure time

    // Property: ExposureTime
    // Parameter: Exposure time of the camera in millisecond. Maximum value is 419ms.
    float f=(float)10.0;
    if(_camera->SetCameraPropertyF("ExposureTime", f) < 0)
        throw std::runtime_error ("Set camera property failed: ExposureTime");

    // 8- Set the other properties

    int iThreshold=210;
    // Property:Peak0_Threshold
    // Peak detector gray-level threshold (pre-filter threshold)
    // (Min:0 Max:2047)
    if(_camera->SetCameraPropertyI("Peak0_Threshold", iThreshold) < 0)
        throw std::runtime_error ("Set camera property failed: Peak0_Threshold");
    // Property:Peak1_Threshold
    // Peak detector gray-level threshold (pre-filter threshold)
    // (Min:0 Max:2047)
    if(_camera->SetCameraPropertyI("Peak1_Threshold", iThreshold) < 0)
        throw std::runtime_error ("Set camera property failed: Peak1_Threshold");


    // Feel free to add some more properties...

    // Read the properties to check that they are well configured
    _camera->GetCameraPropertyI("PeakDetector.Mode",&iProperty);
    _camera->GetCameraPropertyI("Width",&iProperty);
    _camera->GetCameraPropertyI("Height",&iProperty);
    _camera->GetCameraPropertyI("Trigger.Source",&iProperty);
    _camera->GetCameraPropertyF("ExposureTime",&f);

    return 0;	// No error
}

int capture(pf3dcam::Camera *_camera)
{
    int iProfilecount;
    bool allocbuffer=true;
    unsigned char *imgBuf;
    char text[128];

    // Prepare the file stream to store the 3D data in a txt file
    string strPathPeakInfo(gsFullAppPath);
    #ifdef WIN32
    strPathPeakInfo.append("\\PeakInfo.txt");
    #endif
    #ifdef _UNIX_
    strPathPeakInfo.append("/PeakInfo.txt");
    #endif
    ofstream info_file(strPathPeakInfo.c_str());
    // Prepare the file stream to store the status info of each frame in a txt file
    string strPathStatusInfo(gsFullAppPath);
    #ifdef WIN32
    strPathStatusInfo.append("\\StatusInfo.txt");
    #endif
    #ifdef _UNIX_
    strPathStatusInfo.append("/StatusInfo.txt");
    #endif
    ofstream status_file(strPathStatusInfo.c_str());

    try
    {
        // Start continous grabbing
        _camera->grab ();
        cout << "Grabbing..." << endl;

        iProfilecount=0;
        // Grab profiles until MAXPROFILES reached or a timeout occurs
        while(iProfilecount++ < MAXPROFILES)
        {
            try
            {
                // Wait until a new frame is available or TIMEOUT is reached
                _camera->wait (TIMEOUT);

                // Get the Raw Frame
                sal3d::Frame rawFrame = _camera->rawFrame();

                // Get the 3D data, , which is available in PeakDetector.Mode = 2D+3D or 3Donly
                // The has3DInfo() function checks if the PeakDetector.Mode is 3Donly or 2D+3D.
                // If it returns true, that means that the frame contains 3D data.
                if(_camera->has3DInfo()) 
                {
                    sal3d::Frame::Plane<unsigned char> p(rawFrame, 0);
                    // Status information variables 
                    unsigned int ImageCounter = 0;       // 3 bytes
                    unsigned int RealTimeCounter = 0;    // 4 bytes
                    unsigned int EncoderPosition = 0;    // 3 bytes
                    unsigned int MissedTriggerCounter = 0;   // 1 byte
                    unsigned int MissedBurstTriggerCounter = 0;   // 1 byte
                    unsigned int MissedFCTriggerCounter = 0;   // 1 byte
                    unsigned char halfbyteorder = 0;
                    unsigned int stat = 0;

                    // For each pixel of the profile, get the 3D information and status
                    for(int x=0; x < p.width(); x++)
                    {
                        // 3D Data Block: For each column there are 5 lines that contain the 3D data.
                        // More information in the section "3D data format" of the 3D03 camera manual
                        // Row 0: 2D Image information
                        // Row 1: Peak[15:8]
                        // Row 2: Peak[7:0] -> Note: Peak[15:6] Integer part: Peak[5:0] Fractional Part
                        // Row 3: LL_WIDTH [5:0] Laser Line Width
                        // Row 4: LL_HEIGHT [8:4] Laser Line Height; STAT[3:0] Status information
                        int y=p.height()-4;
                        // Get the peak value in float format
                        float fPeak = ((p.pixel(x,y)<<8) + p.pixel(x,y+1))/64;
                        // Get the peak value in 8bit format
                        unsigned char cPeak = p.pixel(x,y);
                        // Get the width
                        unsigned char cWidth = p.pixel(x,y+2);
                        // Get the width
                        unsigned char cQuality = (p.pixel(x,y+3))>>4;
                        // Save the 3D data into a txt file
                        sprintf(text,"Profile:%d Col:%d Peak(float):%f Peak(8bits):%d Width:%d Quality:%d",
                            iProfilecount, x, fPeak, cPeak, cWidth, cQuality);
                        info_file << text << endl;
                        //cout << text << endl;
                        
                        // Get the corresponding part of the status information line
                        // More information in the section "Status information" of the 3D03 camera manual
                        if(x<6) //ImageCounter: 3 bytes (6 = 3 * 2 halfbytes)
                        {
                            stat = (unsigned int)(p.pixel(x,y+3) & 0x0f)<<(4*halfbyteorder);
                            ImageCounter = ImageCounter + stat;
                            halfbyteorder++;
                            if(halfbyteorder==6) halfbyteorder = 0;
                        }
                        if(x>=8 && x<16) //RealTimeCounter: 4 bytes
                        {
                            stat = (unsigned int)(p.pixel(x,y+3) & 0x0f)<<(4*halfbyteorder);
                            RealTimeCounter = RealTimeCounter + stat;
                            halfbyteorder++;
                            if(halfbyteorder==8) halfbyteorder = 0;
                        }
                        if(x>=16 && x<22) //EncoderPosition: 3 bytes
                        {
                            stat = (unsigned int)(p.pixel(x,y+3) & 0x0f)<<(4*halfbyteorder);
                            EncoderPosition = EncoderPosition + stat;
                            halfbyteorder++;
                            if(halfbyteorder==6) halfbyteorder = 0;
                        }
                        if(x>=24 && x<26) //MissedTriggerCounter: 1 byte
                        {
                            stat = (unsigned int)(p.pixel(x,y+3) & 0x0f)<<(4*halfbyteorder);
                            MissedTriggerCounter = MissedTriggerCounter + stat;
                            halfbyteorder++;
                            if(halfbyteorder==2) halfbyteorder = 0;
                        }
                        if(x>=32 && x<34) //MissedBurstTriggerCounter: 1 byte
                        {
                            stat = (unsigned int)(p.pixel(x,y+3) & 0x0f)<<(4*halfbyteorder);
                            MissedBurstTriggerCounter = MissedBurstTriggerCounter + stat;
                            halfbyteorder++;
                            if(halfbyteorder==2) halfbyteorder = 0;
                        }
                        if(x>=40 && x<42) //MissedFrameCombineTriggerCounter: 1 byte
                        {
                            stat = (unsigned int)(p.pixel(x,y+3) & 0x0f)<<(4*halfbyteorder);
                            MissedFCTriggerCounter = MissedFCTriggerCounter + stat;
                            halfbyteorder++;
                            if(halfbyteorder==2) halfbyteorder = 0;
                        }
                    }
                    // Save the status info into a file
                    sprintf(text,"Profile:%d ImageCounter:%u RealTimeCounter:%u EncoderPosition:%u",
                            iProfilecount, ImageCounter, RealTimeCounter, EncoderPosition);
                    status_file << text ;
                    sprintf(text," MissedTriggerCounter:%u MissedBurstTriggerCounter:%u MissedFCTriggerCounter:%u",
                            MissedTriggerCounter, MissedBurstTriggerCounter, MissedFCTriggerCounter);
                    status_file << text << endl;
                }
                cout << ".";  
            }
            catch (pf3dcam::Timeout &)
            {
                cout << "Timeout";
            }
            catch (std::exception &error)
            {
                cerr << error.what() << endl;
            }
		}

        // Stop grabbing
        _camera->freeze ();
    } 
    catch (exception& e)
    {
        cerr << "exception caught: " << e.what() << endl;
    }
    cout << endl;

    return 0;
}

int main(int argc, char **argv)
{
    int pfPort = 0;
    pf3dcam::Camera *_camera;
    char ConfigFile[255];


    // Tet the application path and define several path to store data
    _getcwd(gsFullAppPath, MAX_PATH);

    // Open the camera port

    try
    {
        _camera = 0;
        // Select one of the following ways to connect the camera
        // Connect the camera selecting the IP address
        //_camera = new pf3dcam::Camera("|ip=192.168.1.203");

        // Connect the camera selecting the MAC address
        //_camera = new pf3dcam::Camera("|mac=00:11:1c:f5:a0:d3");

        // Connect the camera using a PV configuration file (*.pvcfg)
        // By using this file, there is no need to configure the camera again with the "config" function
        //sprintf(ConfigFile, "configfile=%s/PF3DSuite2/configfiles/PF_3d03.pvcfg", getenv("PF_ROOT"));	
        //_camera = new pf3dcam::Camera(ConfigFile);

        // Connect the camera using a txt configuration file (*.txt)
        //sprintf(ConfigFile, "configfile=%s/PF3DSuite2/configfiles/PF_3d03.txt", getenv("PF_ROOT"));	
        //_camera = new pf3dcam::Camera(ConfigFile);

        // Select the camera through the Finder dialog
        _camera = new pf3dcam::Camera("");
    }
    catch(...) {
        delete _camera;
        cerr << "Problem opening the camera. Press any key to exit the application. " << endl;
        fflush(stdin);
        while(!_kbhit());
        return -1;
    }

    // Configure some properties of the camera
    try
    {
        //Comment the following line when using a PV configuration file (*.pvcfg)
        config(_camera);
    }
    catch(runtime_error &e) {
        delete _camera;
        cerr << "Uups... Problem configuring the camera. Press any key to exit the application. " << endl << e.what();
        fflush(stdin);
        while(!_kbhit());
        return -1;
    }

    // Capture some frames 
    try
    {
        capture(_camera);
        delete _camera;
        cout << "Acquisition finished. Press any key to exit the application." << endl;
        fflush(stdin);
        while(!_kbhit());
        return 0;
    }
    catch (...)
    {
        delete _camera;
        cerr << "Uups... Problem acquiring from the frame grabber. Press any key to exit the application." << endl;
        fflush(stdin);
        while(!_kbhit());
        return -1;
    }
}
