/**
//  \file 3D03-GetRangeMap.cpp
//	
//  \brief
//  This example is a quick start application to get familar with the
//  3D camera MV1-D2048X1088-3D03-760-G2 and the SDK based on the "sal3d" library.
//  This example shows how to configure the camera with a set of properties, 
//  grab some frames, create a range map and store the 2D and 3D information.
//
//  Description: The main functionality of the example is to configure the
//  camera with a set of properties, grab some frames and create a range map.
//  The properties setup is implemented in the "config" function. 
//  The property list is available in the documentation of the PFRemote. 
//  
//  The main purpose of the "capture" function is to grab some frames from 
//  the framegrabber and store the information.
//  Depending on the PeakDetector_Mode property, the information contained in 
//  the frames is different:
//  - In "Mode_2Donly" laser detection is turned off and camera behaves as a 
//  normal area scan camera. In this case, the example saves the frames (images)
//  in .bmp file format into the subfolder "Frames". 
//
//  - In "Mode2Dand3D" laser line detection is turned on. The sensor image (2D image) 
//  is transmitted together with the 3D data. In this case, the example extracts 
//  a 3D profile from each frame (laser peak information) and creates a  Range map. 
//  The 3D data contains also information about the laser width and quality. 
//  Finally the rangemap is saved into an arm file and the laser peak data is stored 
//  into a txt file. The example also saves the frames (images) in .bmp file format 
//  into the subfolder "Frames".
//
//  - In "Mode_3Donly" laser line detection is turned on and only 3D data plus an 
//  additional image row is transmitted. The scan rate of this mode is considerably 
//  faster than the 2D&3D mode. In this case, the example extracts a 3D profile 
//  from each frame (laser peak information) and creates a Range map. 
//  The 3D data contains also information about the laser width and quality. 
//  Finally the rangemap is saved into a arm file and the laser peak data is stored 
//  into a txt file.
//
//  The Range (*.arm), With (*.awm) and Quality (*.aqm) map files can be visualized
//  with the PF3DSuite. And the SAL3D examples allow the user to create, for instance,
//  a COP or a ZMAP from a Range map (*.arm). For more information see the SAL3D 
//  documentation.
// 
*/

#include <iostream>
#include <fstream>

#ifdef WIN32
#include <direct.h>
#include <conio.h>
#endif // WIN32

#ifdef _UNIX_

#include <termios.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <dlfcn.h>
#include <string.h>

int _kbhit(void)
{
  struct termios oldt, newt;
  int ch;
  int oldf;

  tcgetattr(STDIN_FILENO, &oldt);
  newt = oldt;
  newt.c_lflag &= ~(ICANON | ECHO);
  tcsetattr(STDIN_FILENO, TCSANOW, &newt);
  oldf = fcntl(STDIN_FILENO, F_GETFL, 0);
  fcntl(STDIN_FILENO, F_SETFL, oldf | O_NONBLOCK);

  ch = getchar();

  tcsetattr(STDIN_FILENO, TCSANOW, &oldt);
  fcntl(STDIN_FILENO, F_SETFL, oldf);

  if(ch != EOF)
  {
    ungetc(ch, stdin);
    return 1;
  }

  return 0;
}

#define _getch getchar
#define _getcwd getcwd
#define MAX_PATH 256

#endif // _UNIX_

#include <sal3dpp/Frame.hpp>
#include <sal3dpp/RangeMap.hpp>
#include <sal3dpp/Profile.hpp>

#include <pf3dcam/Camera.hpp>

/// \brief
/// Maximum number of frames (profiles) to be captured
///
#define MAXPROFILES 200

/// \brief
/// Timeout (in milliseconds) to receive a valid frame from the camera
///
#define TIMEOUT 1000

#define CURRENTFRAME true
#define GETNEWFRAME false

using namespace std;
using namespace sal3d;
using namespace pf3dcam;

namespace
{
    char gsFullAppPath[MAX_PATH];
}

///
/// \brief To configure some properties of the camera
///
/// \return Error
///
int config(pf3dcam::Camera *_camera)
{
    //Inicialization for 3D03 camera only

    // Check the camera type
    CAM_Type _camType=CAM_UNKNOWN;
    _camType = _camera->type();
    if (_camType!=CAM_3D03)
        throw std::runtime_error ("This is not a 3D03 camera. This example is only valid for 3D03 cameras");

    // Setup the camera configuration

    // 1- Set the Width and OffsetX properties
    // Property: Width - equivalent to Window.W
    // Parameter: Width of the Window 
    // (Min:32 Max:2048, in increments of 8 pixels)
    if(_camera->SetCameraPropertyI("Width", 1600) < 0)
        throw std::runtime_error ("Set camera property failed: Width");
    // Property: OffsetX - equivalent to Window.X
    // (Min:0 Max:448, in increments of 8 pixels)
    if(_camera->SetCameraPropertyI("OffsetX", 200) < 0)
        throw std::runtime_error ("Set camera property failed: OffsetX");

    // 2- Set the Peak0 ROI
    // Property: Peak0_3DY- equivalent to Window.Y
    // (Min:0 Max:1072, in increments of 1 pixel)
    if(_camera->SetCameraPropertyI("Peak0_3DY", 100) < 0)
        throw std::runtime_error ("Set camera property failed: Peak0_3DY");
    // Property: Peak0_3DH- equivalent to Window.H
    // (Min:16 Max:1009, in increments of 1 pixel)
    if(_camera->SetCameraPropertyI("Peak0_3DH", 1000) < 0)
        throw std::runtime_error ("Set camera property failed: Peak0_3DH");

    // 3- Set the number of peaks. In this case only 1
    // Property: PeakDetector_NrOfPeaks (1 or 2)
    if(_camera->SetCameraPropertyI("PeakDetector_NrOfPeaks", 1) < 0)
        throw std::runtime_error ("Set camera property failed: PeakDetector_NrOfPeaks");

    // 4- Set the 3D mode
    // Property: PeakDetector.Mode	
    // Parameter: 0: "Img2D" (2D only)
    //            1: "Img2D3D" (2D+3D)
    //            2: "Img3D" (3D only)
    if(_camera->SetCameraPropertyI("PeakDetector.Mode", 1) < 0)
        throw std::runtime_error ("Set camera property failed: PeakDetector.Mode");

    // 5- Read the HeightInterface property and set the Height property according to it
    int iProperty;
    _camera->GetCameraPropertyI("HeightInterface",&iProperty);
    if(_camera->SetCameraPropertyI("Height", iProperty) < 0)
        throw std::runtime_error ("Set camera property failed: Height");

    // 6- Set the Trigger properties

    // Property: Trigger.Source
    // Parameter: 0 : Free_Running (Trigger is on Free Running Mode)
    //            1 : Interface_Trigger (Interface Trigger is the trigger source)
    //            2 : IO_Trigger (I/O Trigger is the trigger source)
    if(_camera->SetCameraPropertyI("Trigger.Source", 0) < 0)
        throw std::runtime_error ("Set camera property failed: Trigger.Source");

    // Property: TriggerDivider
    // Parameter: TriggerDivider value - only take every n-th trigger (n=TriggerDivider)
    // TriggerDivider=1 means: take every trigger (default behaviour)
    if(_camera->SetCameraPropertyI("TriggerDivider", 1) < 0)
        throw std::runtime_error ("Set camera property failed: TriggerDivider");

    // 7- Set the exposure time

    // Property: ExposureTime
    // Parameter: Exposure time of the camera in millisecond. Maximum value is 419ms.
    float f=(float)10.0;
    if(_camera->SetCameraPropertyF("ExposureTime", f) < 0)
        throw std::runtime_error ("Set camera property failed: ExposureTime");

    // 8- Set the other properties

    int iThreshold=210;
    // Property:Peak0_Threshold
    // Peak detector gray-level threshold (pre-filter threshold)
    // (Min:0 Max:2047)
    if(_camera->SetCameraPropertyI("Peak0_Threshold", iThreshold) < 0)
        throw std::runtime_error ("Set camera property failed: Peak0_Threshold");
    // Property:Peak1_Threshold
    // Peak detector gray-level threshold (pre-filter threshold)
    // (Min:0 Max:2047)
    if(_camera->SetCameraPropertyI("Peak1_Threshold", iThreshold) < 0)
        throw std::runtime_error ("Set camera property failed: Peak1_Threshold");

    
    // Feel free to add some more properties...

    // Read the properties to check that they are well configured
    _camera->GetCameraPropertyI("PeakDetector.Mode",&iProperty);
    _camera->GetCameraPropertyI("Width",&iProperty);
    _camera->GetCameraPropertyI("Height",&iProperty);
    _camera->GetCameraPropertyI("Trigger.Source",&iProperty);
    _camera->GetCameraPropertyF("ExposureTime",&f);

    return 0;	// No error
}

int capture(pf3dcam::Camera *_camera)
{
    char sFramePathAndFile[MAX_PATH];
    int iProfilecount;
    int _profilesLength;
    int mode;

    // In the 3D03 the profile length corresponds to the window width
    _profilesLength = _camera->profileLength();
    // Create the Range, Width and Quality maps
    sal3d::RangeMap rm(_profilesLength,MAXPROFILES);
    sal3d::RangeMap wm(_profilesLength,MAXPROFILES);
    sal3d::RangeMap qm(_profilesLength,MAXPROFILES);

    // Get the camera mode (2Donly, 2D+3D or 3Donly) in order
    // to know which information is available in the frame
    _camera->GetCameraPropertyI("PeakDetector.Mode",&mode);

    try
    {
        // Start continous grabbing
        _camera->grab ();
        cout << "Grabbing..." << endl;

        iProfilecount=0;
        // Add profiles to the rangemap until MAXPROFILES reached or a timeout occurs
        while(iProfilecount++ < MAXPROFILES)
        {
            try
            {
                // Wait until a new frame is available or TIMEOUT is reached
                _camera->wait (TIMEOUT);

                if(mode==0) // 2Donly Mode
                {
                    // Get the raw image data
                    sal3d::Frame rawFrame = _camera->rawFrame(GETNEWFRAME);
                    // Save the frames in .bmp file format into the subfolder "Frames" (make sure the folder exists)
                    #ifdef WIN32
                    sprintf_s(sFramePathAndFile, MAX_PATH, "%s\\Frames\\frame_%04d.bmp", gsFullAppPath, iProfilecount);
                    #endif
                    #ifdef _UNIX_
                    sprintf(sFramePathAndFile, "%s/Frames/frame_%04d.bmp", gsFullAppPath, iProfilecount);
                    #endif
                    // Uncomment in case you wants to use this function.
                    // Please consider this could slow down your performance in grabbing frames.
                    rawFrame.saveToFile(sFramePathAndFile);
                }
                else if(mode==1) // 2D+3D Mode
                {
                    // Get profile (peak), width and quality data
                    sal3d::Profile p(rm.getNewProfile());
                    sal3d::Profile wp(wm.getNewProfile());
                    sal3d::Profile qp(qm.getNewProfile());
                    float *wData = wp.begin();
                    float *qData = qp.begin();
                    _camera->profile(p, GETNEWFRAME);
                    _camera->width(wData, _profilesLength, CURRENTFRAME);
                    _camera->quality(qData, _profilesLength, CURRENTFRAME);
                    // Get the 2D image data from the frame
                    sal3d::Frame f = _camera->frame(CURRENTFRAME);
                    #ifdef WIN32
                    // Save the frames in .bmp file format into the subfolder "Frames" (make sure the folder exists)
                    sprintf_s(sFramePathAndFile, MAX_PATH, "%s\\Frames\\frame_%04d.bmp", gsFullAppPath, iProfilecount);
                    #endif
                    #ifdef _UNIX_
                    sprintf(sFramePathAndFile, "%s/Frames/frame_%04d.bmp", gsFullAppPath, iProfilecount);
                    #endif
                    // Uncomment in case you wants to use this function.
                    // Please consider this could slow down your performance in grabbing frames.
                    f.saveToFile(sFramePathAndFile);
                }
                else // 3Donly Mode
                {
                    // Get profile (peak), width and quality data
                    sal3d::Profile p(rm.getNewProfile());
                    sal3d::Profile wp(wm.getNewProfile());
                    sal3d::Profile qp(qm.getNewProfile());
                    float *wData = wp.begin();
                    float *qData = qp.begin();
                    _camera->profile(p, GETNEWFRAME);
                    _camera->width(wData, _profilesLength, CURRENTFRAME);
                    _camera->quality(qData, _profilesLength, CURRENTFRAME);
                }

                cout << ".";        
            }
            catch (pf3dcam::Timeout &)
            {
                cout << "Timeout";
            }
            catch (std::exception &error)
            {
                cerr << error.what() << endl;
            }
        }

        // Stop grabbing
        _camera->freeze ();
    } 
    catch (exception& e)
    {
        cerr << "exception caught: " << e.what() << endl;
    }
    cout << endl;

    iProfilecount = rm.countProfiles();
    cout << "Profiles in Rangemap rm1: " << iProfilecount << endl;

    // In case there are 3D profiles, save them to a rangemap file or txt file
    if (iProfilecount > 0){
        // Create the Range, Width and Quality map files
        string strRangeMapFile(gsFullAppPath);
        string strWidthMapFile(gsFullAppPath);
        string strQualityMapFile(gsFullAppPath);
        #ifdef WIN32
        strRangeMapFile.append("\\rangemap.arm");
        strWidthMapFile.append("\\rangemap.awm");
        strQualityMapFile.append("\\rangemap.aqm");
        #endif
        #ifdef _UNIX_
        strRangeMapFile.append("/rangemap.arm");
        strWidthMapFile.append("/rangemap.awm");
        strQualityMapFile.append("/rangemap.aqm");
        #endif

        cout << "Saving Rangemap and peak data to a file:" << endl;
        rm.saveToFile(strRangeMapFile);
        wm.saveToFile(strWidthMapFile);
        qm.saveToFile(strQualityMapFile);

        // Save the peak, width and quality from the profile
        string strPathPeak(gsFullAppPath);
        string strPathInfo(gsFullAppPath);

        #ifdef WIN32
        strPathInfo.append("\\info.txt");
        strPathPeak.append("\\peak.txt");
        #endif
        #ifdef _UNIX_
        strPathInfo.append("/info.txt");
        strPathPeak.append("/peak.txt");
        #endif

        ofstream info_file(strPathInfo.c_str());
        ofstream peak_file(strPathPeak.c_str());

        for(int i=0; i < iProfilecount; i++){
            // Get the profile from the range map
            sal3d::Profile profile = rm[i];
            sal3d::Profile width = wm[i];
            sal3d::Profile quality = qm[i];
            cout << ".";
            for(size_t r=0; r < profile.length(); r++){
	            //float f = profile[r];
	            //cout << "Profile(" << i << ") Peak : Width : Quality (" << r << ") - " << profile[r];
	            //cout << " : " << width[i][r] << " : " << quality[i][r] << endl;
            	
	            // Save peak, quality and width data for each profile in a file 
	            peak_file << profile[r] << " ";

	            info_file << "Profile(" << i << ") Row(" << r << ") - Peak: " << profile[r];
                info_file << ", Width : " << width[r] << ", Quality: " << quality[r] << endl;
            }
            peak_file << endl;
            info_file << endl;
        }
        cout << "." << endl;
    }
    
    return 0;
}

int main(int argc, char **argv)
{
    int pfPort = 0;
    pf3dcam::Camera *_camera;
    char ConfigFile[255];


    // Tet the application path and define several path to store data
    _getcwd(gsFullAppPath, MAX_PATH);

    // Open the camera port

    try
    {
        _camera = 0;
        // Select one of the following ways to connect the camera
        // Connect the camera selecting the IP address
        //_camera = new pf3dcam::Camera("|ip=192.168.1.140");

        // Connect the camera selecting the MAC address
        //_camera = new pf3dcam::Camera("|mac=00:11:1c:f5:a0:d3");

        // Connect the camera using a PV configuration file (*.pvcfg)
        // By using this file, there is no need to configure the camera again with the "config" function
        //sprintf(ConfigFile, "configfile=%s/PF3DSuite2/configfiles/PF_3d03.pvcfg", getenv("PF_ROOT"));	
        //_camera = new pf3dcam::Camera(ConfigFile);

        // Connect the camera using a txt configuration file (*.txt)
        //sprintf(ConfigFile, "configfile=%s/PF3DSuite2/configfiles/PF_3d03.txt", getenv("PF_ROOT"));	
        //_camera = new pf3dcam::Camera(ConfigFile);

        // Select the camera through the Finder dialog
        _camera = new pf3dcam::Camera("");
    }
    catch(...) {
        delete _camera;
        cerr << "Problem opening the camera. Press any key to exit the application. " << endl;
        fflush(stdin);
        while(!_kbhit());
        return -1;
    }
    
    // Configure some properties of the camera
    try
    {
        //Comment the following line when using a PV configuration file (*.pvcfg)
        config(_camera);
    }
    catch(runtime_error &e) {
        delete _camera;
        cerr << "Uups... Problem configuring the camera. Press any key to exit the application. " << endl << e.what();
        fflush(stdin);
        while(!_kbhit());
        return -1;
    }
	
    // Capture some frames 
    try
    {
        capture(_camera);
        delete _camera;
        cout << "Acquisition finished. Press any key to exit the application." << endl;
        fflush(stdin);
        while(!_kbhit());
        return 0;
    }
    catch (...)
    {
        delete _camera;
        cerr << "Uups... Problem acquiring from the frame grabber. Press any key to exit the application." << endl;
        fflush(stdin);
        while(!_kbhit());
        return -1;
    }
}
