/*
 * Camera class from pf3dcamera
 * Copyright (C) 2011 AQSENSE, S.L.
 */
#if !defined (CAMERA_HPP)
#define CAMERA_HPP

#if defined (_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif // _MSC_VER && _MSC_VER >= 1020

#include <stdexcept>
#include <string>
#include <cstdio>
#include <sal3dpp/Frame.hpp>
#include <sal3dpp/Profile.hpp>
#ifdef WIN32
#include <pfcam.h>
#endif
#include <pf3dcamC/camera.h>
#include "Error.hpp"

namespace pf3dcam
{

inline int
GetNumPorts()
{
    int nPorts;
    Error e;
    pf3dcam_get_num_ports(&nPorts, &e.e);
    e.throw_if_needed();
    return nPorts;
}

inline void
OpenPort(int port)
{
    Error e;
    pf3dcam_open_port(port, &e.e);
    e.throw_if_needed();
}

inline void
ClosePort(int port)
{
    Error e;
    pf3dcam_close_port(port, &e.e);
    e.throw_if_needed();
}

inline int GetDllVersion()
{
    return pf3dcam_get_dll_version();
}

///
/// \class Camera
/// \brief An instance of an acquisition device
///
/// The Camera class is the interface to access an acquisition device, usually
/// consisting of a frame grabber and the connected camera.
///
/// There are two constructors which can identify the device for its port
/// or the frame grabber type, and the board and channel number. There also
/// exists a 3rd constructor which should only be used for the file frame
/// grabber, which is useful for analyzing frames from file.
///
class Camera
{
    public:
        ///
        /// \brief Constructor from port. This constructs a Camera from a
        /// specific port id, as they are enumerated in pflib.
        ///
        /// \param[in] port The camera port.
        /// \param[in] config A string with the frame grabber configuration.
        ///
        /// \throw std::runtime_error() if there is an error creating the
        ///        Camera.
        ///
        Camera(int port, const std::string &config = std::string("")):
            _port(port)
        {
            _camera.p = 0;
            Error e;
            pf3dcam_camera_new_port(port, config.c_str(),
                    &_camera, &e.e);
            e.throw_if_needed();
        }

        ///
        /// \brief Constructor from a specific frame grabber board and channel.
        ///
        /// \param[in] type The type of the frame grabber.
        /// \param[in] board The board number.
        /// \param[in] channel The channel number in the board.
        /// \param[in] config A string with the frame grabber configuration.
        ///
        /// \throw std::runtime_error() if there is an error creating the
        ///        Camera.
        ///
        Camera(FG_Type type, int board, int channel,
                const std::string &config = std::string(""))
        {
            _camera.p = 0;
            Error e;
            pf3dcam_camera_new_cameralink(type, board, channel,
                    config.c_str(), &_camera, &e.e);
            e.throw_if_needed();
            pf3dcam_camera_get_pflib_port(_camera, &_port, &e.e);
            e.throw_if_needed();
        }

        ///
        /// \brief Constructor for the file frame grabber.
        ///
        /// \warning This constructor should only be used for the file frame
        ///          grabber.
        ///
        /// \param[in] width The width of the image
        /// \param[in] height The height of the image
        /// \param[in] config A string with the frame grabber configuration.
        ///
        /// \throw std::runtime_error() if there is an error creating the
        ///        Camera.
        ///
        Camera(int width, int height,
                const std::string &config = std::string("")):
            _port(-1)
        {
            _camera.p = 0;
            Error e;
            pf3dcam_camera_new_file(width, height, config.c_str(),
                    &_camera, &e.e);
            e.throw_if_needed();
        }

        ///
        /// \brief Constructor for the gige cameras
        ///
        /// \param[in] config A string to identify the camera by IP or MAC.
        ///
        /// \throw std::runtime_error() if there is an error creating the
        ///        Camera.
        ///
        Camera(const std::string &config = std::string("")):
            _port(-1)
        {
            _camera.p = 0;
            Error e;
            pf3dcam_camera_new_gige(config.c_str(), &_camera, &e.e);
            e.throw_if_needed();
        }

        ///
        /// \brief Destructor.
        ///
        /// If the reference counter reaches zero, the object is destroyed
        /// and its memory released.
        ///
        ~Camera()
        {
            pf3dcam_camera_release(_camera);
        }

        ///
        /// \brief Copy constructor.
        ///
        /// \param[in] cam The Camera to copy from.
        ///
        Camera(const Camera &cam):
            _camera(cam._camera),
            _port(cam._port)
        {
            Error e;
            pf3dcam_camera_share(_camera, &e.e);
            e.throw_if_needed();
        }

        ///
        /// \brief Assignment operator.
        ///
        /// \param[in] cam The Camera to copy from.
        ///
        /// \return A reference to this object.
        ///
        Camera &
        operator=(const Camera &cam)
        {
            pf3dcam_camera_release(_camera);
            _camera = cam._camera;
            pf3dcam_camera_share(_camera);
            _port = cam._port;

            return *this;
        }

        ///
        /// \brief Get the pflib port number associated with the Camera.
        ///
        /// \return The port id of the Camera.
        ///
        int
        getPFLibPort() const
        {
            return _port;
        }

        CAM_Type
        type() const
        {
            CAM_Type type;
            Error e;
            pf3dcam_camera_get_type(_camera, &type, &e.e);
            e.throw_if_needed();
            return type;
        }

        CAM_Iface
        interface() const
        {
            CAM_Iface iface;
            Error e;
            pf3dcam_camera_get_iface(_camera, &iface, &e.e);
            e.throw_if_needed();
            return iface;
        }

        ///
        /// \brief Synchronize camera configuration
        ///
        /// For a cameralink camera, this updates the frame grabber config
        /// according to the camera configuration.
        ///
        /// For a gige camera, this rereads the configuration from the camera
        /// in case it's been changed through the dialogs.
        ///
        /// \throw std::runtime_error() if there is an unknown error.
        ///
        void
        syncConfig() const
        {
            Error e;
            pf3dcam_camera_sync_config(_camera, &e.e);
            e.throw_if_needed();
        }

        void
        grab()
        {
            Error e;
            pf3dcam_camera_grab(_camera, &e.e);
            e.throw_if_needed();
        }

        void
        freeze()
        {
            Error e;
            pf3dcam_camera_freeze(_camera, &e.e);
            e.throw_if_needed();
        }

        void
        wait(int timeout = -1)
        {
            Error e;
            pf3dcam_camera_wait(_camera, timeout, &e.e);
            e.throw_if_needed();
        }

        ///
        /// \brief Get an acquired raw Frame.
        ///
        /// This function returns a raw Frame acquired from the camera. Raw
        /// meaning that it contains the image but also the 3D information, if
        /// available.
        ///
        /// By default (\p getCurrent = false) this function takes the next
        /// acquired Frame and returns it. If \p getCurrent is set to true, the
        /// returned Frame is the last captured, no new one is taken.
        ///
        /// \param[in] getCurrent Whether to grab a new Frame.
        ///
        /// \return The acquired Frame.
        ///
        /// \throw std::runtime_error() if there is an error creating the
        ///        Camera.
        ///
        sal3d::Frame
        rawFrame(bool getCurrent = false)
        {
            sal3d_frame f;
            Error e;
            pf3dcam_camera_get_raw_frame(_camera, getCurrent, &f, &e.e);
            e.throw_if_needed();

            return sal3d::Frame(f);
        }

        ///
        /// \brief Get an acquired Frame.
        ///
        /// This function returns a Frame acquired from the camera.
        ///
        /// By default (\p getCurrent = false) this function takes the next
        /// acquired Frame and returns it. If \p getCurrent is set to true, the
        /// returned Frame is the last captured, no new one is taken.
        ///
        /// \param[in] getCurrent Whether to grab a new Frame.
        ///
        /// \return The acquired Frame.
        ///
        /// \throw std::runtime_error() if there is an error creating the
        ///        Camera.
        ///
        sal3d::Frame
        frame(bool getCurrent = false)
        {
            sal3d_frame f;
            Error e;
            pf3dcam_camera_get_frame(_camera, getCurrent, &f, &e.e);
            e.throw_if_needed();

            return sal3d::Frame(f);
        }

        ///
        /// \brief Get the length of the detected Profile.
        ///
        /// This function returns the length of the detected Profile.
        ///
        /// \return The length of the detected Profile.
        ///
        /// \throw std::runtime_error() if there is an error creating the
        ///        Camera.
        ///
        int
        profileLength() const
        {
            int ret;
            Error e;
            ret = pf3dcam_camera_get_profile_length(_camera, &e.e);
            e.throw_if_needed();

            return ret;
        }

        ///
        /// \brief Get the detected profile.
        ///
        /// This function fills the given Profile with the peak information
        /// detected from the Frame.
        ///
        /// By default (\p getCurrent = false) this function takes the next
        /// acquired Frame and extracts the peak from it. If \p getCurrent is
        /// set to true, the returned Profile is the last captured, no new one
        /// is taken.
        ///
        /// \param[out] p The Profile that will be filled with the peak
        ///             information.
        /// \param[in] getCurrent Whether to grab a new Frame.
        ///
        /// \throw std::runtime_error() if there is an unknown error.
        ///
        void
        profile(sal3d::Profile &p, bool getCurrent = false, int peak = 0)
        {
            Error e;
            pf3dcam_camera_get_profile(_camera, &p[0],
                    p.length(), getCurrent, peak, &e.e);
            e.throw_if_needed();
        }

        ///
        /// \brief Get the detected profile, together with the width and
        /// quality information.
        ///
        /// This function fills the given Profile with the peak information
        /// detected from the Frame, as well as the width of the detected peak
        /// and the quality.
        ///
        /// By default (\p getCurrent = false) this function takes the next
        /// acquired Frame and extracts the peak from it. If \p getCurrent is
        /// set to true, the returned Profile is the last captured, no new one
        /// is taken.
        ///
        /// \param[out] p The Profile that will be filled with the peak
        ///             information.
        /// \param[out] width The width information of the detected peak.
        /// \param[out] quality The quality information of the detected peak.
        /// \param[in] getCurrent Whether to grab a new Frame.
        ///
        /// \throw std::runtime_error() if there is an unknown error.
        ///
        void
        width(std::vector<float> &width, bool getCurrent = false, int peak = 0)
        {
            width.resize(profileLength());
            this->width(&width[0], width.size(), getCurrent, peak);
        }

        void
        width(float *width, int plen, bool getCurrent = false, int peak = 0)
        {
            Error e;
            pf3dcam_camera_get_width(_camera, width,
                    plen, getCurrent, peak, &e.e);
            e.throw_if_needed();
        }

        void
        quality(std::vector<float> &quality, bool getCurrent = false, int peak = 0)
        {
            quality.resize(profileLength());
            this->quality(&quality[0], quality.size(), getCurrent, peak);
        }

        void
        quality(float *quality, int plen, bool getCurrent = false, int peak = 0)
        {
            Error e;
            pf3dcam_camera_get_quality(_camera, quality,
                    plen, getCurrent, peak, &e.e);
            e.throw_if_needed();
        }

        void
        linescan(std::vector<short> &gray, bool getCurrent = false, int peak = 0)
        {
            gray.resize(profileLength());
            linescan(&gray[0], gray.size(), getCurrent, peak);
        }

        void
        linescan(short *gray, int plen, bool getCurrent = false, int peak = 0)
        {
            Error e;
            pf3dcam_camera_get_linescan(_camera, gray,
                    plen, getCurrent, peak, &e.e);
            e.throw_if_needed();
        }

        int
        GetCameraPropertyS(const char* pProperty, char* pValue, int iSize) const
        {
            Error e;
            int ret = pf3dcam_camera_get_property_s(_camera, pProperty, pValue,
                    iSize, &e.e);
            e.throw_if_needed();
            return ret;
        }

        int
        GetCameraPropertyI(const char *pProperty, int *iValue) const
        {
            Error e;
            int ret = pf3dcam_camera_get_property_i(_camera, pProperty, iValue,
                    &e.e);
            e.throw_if_needed();
            return ret;
        }

        int
        GetCameraPropertyInt(const char *pProperty) const
        {
            int val;
            Error e;
            pf3dcam_camera_get_property_i(_camera,
                    pProperty, &val, &e.e);
            e.throw_if_needed();
            return val;
        }

        int
        GetCameraPropertyF(const char* pProperty, float* fValue) const
        {
            Error e;
            int ret = pf3dcam_camera_get_property_f(_camera, pProperty, fValue,
                    &e.e);
            e.throw_if_needed();
            return ret;
        }

        float
        GetCameraPropertyFloat(const char* pProperty, float* fValue) const
        {
            float val;
            Error e;
            pf3dcam_camera_get_property_f(_camera, pProperty, &val, &e.e);
            e.throw_if_needed();
            return val;
        }

        int
        SetCameraPropertyS(const char* pProperty, const char* pValue) const
        {
            Error e;
            int ret = pf3dcam_camera_set_property_s(_camera, pProperty, pValue,
                    &e.e);
            e.throw_if_needed();
            return ret;
        }

        int
        SetCameraPropertyI(const char* pProperty, int iValue) const
        {
            Error e;
            int ret = pf3dcam_camera_set_property_i(_camera, pProperty, iValue,
                    &e.e);
            e.throw_if_needed();
            return ret;
        }

        int
        SetCameraPropertyF(const char* pProperty, float fValue) const
        {
            Error e;
            int ret = pf3dcam_camera_set_property_f(_camera, pProperty, fValue,
                    &e.e);
            e.throw_if_needed();
            return ret;
        }

        void
        showWindow(GIGE_Win win) const
        {
            Error e;
            pf3dcam_camera_show_window(_camera, win, &e.e);
            e.throw_if_needed();
        }

        int
        saveConfigFile(const char* configFile) const
        {
            Error e;
            int ret = pf3dcam_camera_save_config_file(_camera, configFile, &e.e);
            e.throw_if_needed();
            return ret;
        }

	int
        lostCount() const
        {
            Error e;
            int ret = pf3dcam_camera_lost_count(_camera, &e.e);
            e.throw_if_needed();
            return ret;
        }

        bool
        hasFrame()
        {
            //for file fg, we consider 2D3D mode
            if (interface() == CAM_IFACE_UNKNOWN)
                return true;

            int mode;
            mode = GetCameraPropertyInt("PeakDetector.Mode");

            if (mode == 0 || mode == 1)
                return true;
            return false;
        }

        ///
        /// \brief Tell whether the Camera has 3D info available.
        ///
        /// This function returns true if the Camera is configured in any of
        /// the modes with 3D information.
        ///
        /// \return True if the Camera has 3D information available.
        ///
        /// \throw std::runtime_error() if there is an unknown error.
        ///
        bool
        has3DInfo()
        {
            //for file fg, we consider 2D3D mode
            if (interface() == CAM_IFACE_UNKNOWN)
                return true;

            int mode;
            mode = GetCameraPropertyInt("PeakDetector.Mode");

            if (mode == 1 || mode == 2)
                return true;
            return false;
        }

    private:
        pf3dcam_camera _camera;
        int _port;
};
}

#endif //CAMERA_HPP
